#include <jo/jo.h>
#include "modelLoader.h"
#include "mathex.h"

/* Speed of moving */
#define MOVE_SPEED JO_FIXED_1

/* Speed of turning */
#define TURN_SPEED (1319)

/* Number of meshes inside mesh group */
static int IslandMeshCount = 0;

/* Island 3D mesh group */
static jo_3d_mesh * Island;

/* RGB0 plane palette */
static jo_palette Rgb0Palette;

/* Camera location and orientation */
static ExCoordinateSystem Camera = { { 0, -JO_FIXED_120, JO_FIXED_32 }, { JO_FIXED_1, 0, 0 }, { 0, 0, -JO_FIXED_1 } };

/** @brief Move camera forward and backward
 *  @param value Speed of movement
 */
void DoThrust(jo_fixed value)
{
    jo_vector_fixed cross = ExVectorCross(&Camera.AxisX, &Camera.AxisY);
    ExVectorNormal(&cross);

    jo_vector_fixed thrustVector;
    jo_vector_fixed_muls(&cross, value, &thrustVector);
    ExTranslatePoint(&Camera.Origin, &thrustVector);
}

/** @brief Rotate camera around its Up vector
 *  @param value Speed of movement
 */
void DoYaw(jo_fixed value)
{
    Camera.AxisX = ExRotateVectorAroundAxis(&Camera.AxisX, value, &Camera.AxisY);
}

/** @brief Rotate camera around its side vector
 *  @param value Speed of movement
 */
void DoPitch(jo_fixed value)
{
    Camera.AxisY = ExRotateVectorAroundAxis(&Camera.AxisY, value, &Camera.AxisX);
}

/** @brief Rotate camera around its direction vector
 *  @param value Speed of movement
 */
void DoRoll(jo_fixed value)
{
    jo_vector_fixed cross = ExVectorCross(&Camera.AxisX, &Camera.AxisY);
    ExVectorNormal(&cross);
    
    Camera.AxisX = ExRotateVectorAroundAxis(&Camera.AxisX, value, &cross);
    Camera.AxisY = ExVectorCross(&cross, &Camera.AxisX);
    ExVectorNormal(&Camera.AxisY);
}

/** @brief Main demo loop
 */
void DemoLogic()
{
    // Read inputs
    if (jo_is_pad1_available())
    {
        if (jo_is_pad1_key_pressed(JO_KEY_A))
        {
            DoThrust(MOVE_SPEED);
        }
        else if (jo_is_pad1_key_pressed(JO_KEY_B))
        {
            DoThrust(-MOVE_SPEED);
        }

        // Yaw
        if (jo_is_pad1_key_pressed(JO_KEY_LEFT))
        {
            DoYaw(TURN_SPEED);
        }
        else if (jo_is_pad1_key_pressed(JO_KEY_RIGHT))
        {
            DoYaw(-TURN_SPEED);
        }
        
        // Pitch
        if (jo_is_pad1_key_pressed(JO_KEY_UP))
        {
            DoPitch(TURN_SPEED);
        }
        else if (jo_is_pad1_key_pressed(JO_KEY_DOWN))
        {
            DoPitch(-TURN_SPEED);
        }
        
        // Roll
        if (jo_is_pad1_key_pressed(JO_KEY_L))
        {
            DoRoll(TURN_SPEED);
        }
        else if (jo_is_pad1_key_pressed(JO_KEY_R))
        {
            DoRoll(-TURN_SPEED);
        }
    }
}

/** @brief Redering loop
 */
void DemoDraw()
{
    MATRIX matrix;
    
    jo_3d_push_matrix();
    {
        // Set 3D matrix
        ExGetMatrixToCoordinateSystem(&Camera, matrix);
        slLoadMatrix(matrix);

        // Draw island
        for (int mesh = 0; mesh < IslandMeshCount; mesh++)
        {
            jo_3d_mesh_draw(&Island[mesh]);
        }
        
        // Draw water
        jo_background_3d_plane_a_draw(true);

    }
    jo_3d_pop_matrix();
    
    // Draw sky
    jo_background_3d_plane_b_draw(false);
}

/** @brief Palette loader for RGB0
 */
static jo_palette *Rgb0PaletteHandling(void)
{
    // We re-create a new palette for each image. It's not optimal and WILL break with different images that don't have same palette but OK for a demo.
    jo_create_palette(&Rgb0Palette);
    return (&Rgb0Palette);
}

/** @brief Load assets
 */
void LoadAssets()
{
    jo_core_tv_off();

    // set palette loading
    jo_set_tga_palette_handling(Rgb0PaletteHandling);

    // Enable RGB0
    jo_enable_background_3d_plane(JO_COLOR_RGB(71,243,255));

    // Load water layer texture
    jo_img_8bits img;
    img.data = JO_NULL;
    jo_tga_8bits_loader(&img, JO_ROOT_DIR, "WATER.TGA", 0);
    jo_background_3d_plane_a_img(&img, Rgb0Palette.id, true, true);
    jo_free_img(&img);

    // Load sky texture
    img.data = JO_NULL;
    jo_tga_8bits_loader(&img, JO_ROOT_DIR, "SKY.TGA", 0);
    jo_background_3d_plane_b_img(&img, Rgb0Palette.id, true, false);
    jo_free_img(&img);

    // Load island model
    Island= ML_LoadMesh("ISLAND.TMF", JO_ROOT_DIR, &IslandMeshCount);

    jo_core_tv_on();
}

/** @brief Application entry point
 */
void jo_main(void)
{
    // Prepare scene
    jo_core_init(JO_COLOR_Black);

    // This makes 8bpp textures work
    *(volatile unsigned char *)0x060FFCD8 = 0x1F;

    // enable reboot combo
    jo_core_set_restart_game_callback(jo_goto_boot_menu);

    // Initialize demo stuff
    LoadAssets();

    // Start demo
    jo_core_add_callback(DemoDraw);
    jo_core_add_callback(DemoLogic);

    jo_core_run();
}
